/*
 * Copyright (c) 2022 SPARQL Anything Contributors @ http://github.com/sparql-anything
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.sparqlanything.xml;

import com.github.sparqlanything.model.FacadeXGraphBuilder;
import com.github.sparqlanything.model.Slice;
import com.github.sparqlanything.model.Slicer;
import com.github.sparqlanything.model.Triplifier;
import com.github.sparqlanything.model.TriplifierHTTPException;
import com.ximpleware.AutoPilot;
import com.ximpleware.EncodingException;
import com.ximpleware.NavException;
import com.ximpleware.ParseException;
import com.ximpleware.VTDGen;
import com.ximpleware.VTDNav;
import com.ximpleware.XPathEvalException;
import com.ximpleware.XPathParseException;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.jena.ext.com.google.common.collect.Sets;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class XMLTriplifier implements Triplifier, Slicer {

	private static final Logger log = LoggerFactory.getLogger(XMLTriplifier.class);

	public static final String PROPERTY_XPATH = "xml.path";

	private VTDNav buildVTDNav(Properties properties) throws TriplifierHTTPException, IOException, ParseException {
		VTDGen vg = new VTDGen();
		byte[] bytes = IOUtils.toByteArray(Triplifier.getInputStream(properties));
		vg.setDoc(bytes);
		// TODO Support namespaces
		vg.parse(false);
		VTDNav vn = vg.getNav();
		return vn;
	}

	private Iterator<Pair<VTDNav,Integer>> evaluateXPaths(VTDNav vn, List<String> xpaths){
		return new Iterator<Pair<VTDNav, Integer>>() {
			int count = 0;
			Iterator<String> xit = xpaths.iterator();
			Pair<VTDNav, Integer> next = null;
			AutoPilot ap = new AutoPilot(vn);
			String xpath = null;
			@Override
			public boolean hasNext() {
				try {
					if (next != null) {
						return true;
					}
					// If XPath is already loaded, move to the next result
					int result = -1;
					if( xpath != null ){
						result = ap.evalXPath();
						if(result == -1){
							// No more results with this XPath
							xpath = null;
						}
					}
					// Look for next xpath with results
					while (xpath == null && xit.hasNext()) {
						// If XPath is available, load it and move to the next result
						xpath = xit.next();
						log.debug("Evaluating XPath: {}", xpath);
						ap.selectXPath(xpath);
						result = ap.evalXPath();
						if(result == -1){
							xpath = null;
						}else{
							// Stop here for now
							break;
						}
					}

					if(result == -1){
						// No more results
						return false;
					}else{
						// Prepare next result
						next = Pair.of(vn, result);
						return true;
					}
				}catch(XPathParseException|XPathEvalException|NavException e){
					throw new RuntimeException(e);
				}
			}

			@Override
			public Pair<VTDNav, Integer> next() {
				Pair<VTDNav, Integer> toReturn = next;
				next = null;
				return toReturn;
			}
		};
	}

	public void transformWithXPath(List<String> xpaths, Properties properties, FacadeXGraphBuilder builder) throws IOException, TriplifierHTTPException {

		String dataSourceId = Triplifier.getRootArgument(properties);
		String root = Triplifier.getRootArgument(properties);

		builder.addRoot(dataSourceId, root);
		try {
			VTDNav vn = buildVTDNav(properties);
			Iterator<String> xit = xpaths.iterator();
			Iterator<Pair<VTDNav,Integer>> it = evaluateXPaths(vn,xpaths);
			int count = 1;
			while(it.hasNext()) {
				Pair<VTDNav,Integer> next = it.next();
				transformFromXPath(next.getKey(), next.getValue(), count, root, dataSourceId, properties, builder);
				count++;
			}
			log.debug("XPath: {} matches", count);

		} catch (NavException | ParseException e){
			log.error("Error while evaluating XPath expression");
			throw new IOException(e);
		}
	}

	public int transformFromXPath(VTDNav vn, int result, int child, String parentId, String dataSourceId, Properties properties, FacadeXGraphBuilder builder) throws NavException {
		log.trace(" -- index: {} type: {}", result, vn.getTokenType(result));
		switch (vn.getTokenType(result)) {
			case VTDNav.TOKEN_STARTING_TAG:
				String tag = null;
				tag = vn.toString(result);

				log.trace(" -- tag: {} ", tag);
				String childId = String.join("", parentId , "/" , Integer.toString(child), ":", tag);
				builder.addContainer(dataSourceId, parentId, child, childId);

				// Attributes
				int attrCount = vn.getAttrCount();
				log.trace(" -- attr count: {}", attrCount);
				int increment = 0;
				if (attrCount > 0) {
					for (int i = result + 1; i <= result + attrCount; i += 2) {
						// Not sure why but sometime attrCount is not reliable
						if(vn.getTokenType(i) != VTDNav.TOKEN_ATTR_NAME){
							break;
						}
						String key =  vn.toString(i);
						String value = vn.toString(i + 1);
						log.trace(" -- attr: {} = {}", key, value);
						builder.addValue(dataSourceId, childId, key, value);
						increment += 2;
					}
				}
				// Get the text
				int t = vn.getText(); // get the index of the text (char data or CDATA)
				if (t != -1) {
					String text = vn.toNormalizedString(t);
					log.trace(" -- text: {}", text);
					builder.addValue(dataSourceId, childId, 1, text);
				}

				// Iterate on Children until complete
				int tokenDepth = vn.getTokenDepth(result);
				int index = result + increment;
				int childc = 1;
				while(true){
					index++;
					int type =  vn.getTokenType(index);
					String s = vn.toString(index);
					int d = vn.getTokenDepth(index);
					// If type is element and depth is not greater than tokenDepth, break!
					if((type == VTDNav.TOKEN_STARTING_TAG && d <=tokenDepth) || (type == VTDNav.TOKEN_STARTING_TAG && s.equals(""))){
						break;
					}
					log.trace( " ...  index: {} depth: {} type: {} string: {}", index, d, type, s);
					index = transformFromXPath(vn, index, childc, childId, dataSourceId, properties, builder);
					childc++;
				}
				return index - 1;
			case VTDNav.TOKEN_ATTR_NAME:
				// Attribute
				String name = vn.toString(result);
				String value = vn.toString(result + 1);
				log.trace("Attribute {} = {}", name, value);
				String attrChildId = String.join("", parentId , "/" , Integer.toString(child), ":", name);
				builder.addContainer(dataSourceId, parentId, child, attrChildId);
				builder.addValue(dataSourceId, attrChildId, name, value);
				return result + 1;
			case VTDNav.TOKEN_ATTR_VAL:
				// Attribute value
				log.trace("Attribute value: {}", vn.toString(result));
				builder.addValue(dataSourceId, parentId, child, vn.toString(result));
				break;
			case VTDNav.TOKEN_CHARACTER_DATA:
				// Text
				String text = vn.toNormalizedString(result);
				log.trace("Text: {}", text);
				builder.addValue(dataSourceId, parentId, child, vn.toString(result));
				break;
			case VTDNav.TOKEN_DEC_ATTR_NAME:
				log.trace("Attribute (dec): {} = {}", vn.toString(result), vn.toString(result + 1));
				return result + 1;
			case VTDNav.TOKEN_DEC_ATTR_VAL:
				log.trace("Attribute value (dec) ", vn.toString(result));
				break;
			default:
				log.warn("Ignored event: {} {}", vn.getTokenType(result), vn.toString(result));
		}
		return result;
	}

	public void transformSAX(Properties properties, FacadeXGraphBuilder builder) throws IOException, TriplifierHTTPException {

		String namespace = Triplifier.getNamespaceArgument(properties);
		String dataSourceId = Triplifier.getRootArgument(properties);
		String root = Triplifier.getRootArgument(properties);

		XMLInputFactory inputFactory = XMLInputFactory.newInstance();
		// TODO allow users to configure XML parser via properties
		inputFactory.setProperty(XMLInputFactory.IS_SUPPORTING_EXTERNAL_ENTITIES, false);
		inputFactory.setProperty(XMLInputFactory.SUPPORT_DTD, false);

		XMLEvent event = null;
		XMLEventReader eventReader;
		//
		Deque<String> stack = new ArrayDeque<String>();
		Map<String, Integer> members = new HashMap<String, Integer>();
		String path = "";
		StringBuilder charBuilder = null;
		//
		try {
			InputStream is = Triplifier.getInputStream(properties);
			eventReader = inputFactory.createXMLEventReader(is);
		} catch (XMLStreamException e) {
			throw new IOException(e);
		}
		boolean isRoot = true;
		while (eventReader.hasNext()) {

			if (event != null && event.isEndElement()) {
				log.trace("element close: {} [{}]", path, stack.size());
				// Remove the last path
				path = path.substring(0, path.lastIndexOf('/'));

				// Collect data if available
				if (charBuilder != null) {
					String value = charBuilder.toString();
					log.trace("collecting char stream: {}", value);
					if (stack.peekLast() != null && value != null && !value.trim().equals("")) {
						String resourceId = stack.peekLast();
						if (!members.containsKey(resourceId)) {
							members.put(resourceId, 0);
						}
						int member = members.get(resourceId) + 1;
						builder.addValue(dataSourceId, resourceId, member, value);
					}
					charBuilder = null;
				}

				// reset current resource
				stack.removeLast();
			}
			try {
				event = eventReader.nextEvent();
			} catch (XMLStreamException e) {
				throw new IOException("Journey interrupted.", e);
			}
			log.trace("event: {}", event);
			if (event.isStartElement()) {
				StartElement se = event.asStartElement();
				String name;
				if (se.getName().getPrefix().equals("")) {
					name = se.getName().getLocalPart();
				} else {
					name = se.getName().getPrefix() + ":" + se.getName().getLocalPart();
				}
				int member = 0;
				if (stack.size() > 0) {
					String parent = stack.peekLast();
					member = members.get(parent) + 1;
					members.put(parent, member);
				}

				if (path.equals("")) {
					path = String.join("", "/", name);
				} else {
					path = String.join("", path, "/", Integer.toString(member), ":", name);
				}
				log.trace("element open: {} [{}]", path, stack.size());

				// XXX Create an RDF resource
				String resourceId = StringUtils.join("", root, path);
				// If this is the root
				if (isRoot) {
					// Add type root
					builder.addRoot(dataSourceId, resourceId);
					isRoot = false;
				}
				try {
					builder.addType(dataSourceId, resourceId, new URI(toIRI(se.getName(), namespace)));
				} catch (URISyntaxException e) {
					throw new IOException(e);
				}
				if (!members.containsKey(resourceId)) {
					members.put(resourceId, 0);
				}
				// Link it with the container membership property
				if (stack.size() > 0) {
					String parent = stack.peekLast();
					builder.addContainer(dataSourceId, parent, member, resourceId);
				}
				// Attributes
				Iterator<Attribute> attributes = se.getAttributes();
				while (attributes.hasNext()) {
					Attribute attribute = (Attribute) attributes.next();
					log.trace("attribute: {}", attribute);
					try {
						builder.addValue(dataSourceId, resourceId, new URI(toIRI(attribute.getName(), namespace)), attribute.getValue());
					} catch (URISyntaxException e) {
						throw new IOException(e);
					}
				}
				stack.add(resourceId);
			} else if (event.isCharacters()) {
				// Characters
				log.trace("character: {}", event);
				if (charBuilder == null) {
					charBuilder = new StringBuilder();
				}
				charBuilder.append(event.asCharacters().getData().trim());
			}
		}
	}

	private String toIRI(QName qname, String namespace) {
		String ns;
		if(qname.getNamespaceURI().equals("")){
			ns = namespace;
		}else{
			if(qname.getNamespaceURI().endsWith("/") || qname.getNamespaceURI().endsWith("#") ){
				ns = qname.getNamespaceURI();
			}else{
				ns = qname.getNamespaceURI() + '#';
			}
		}
		return ns + qname.getLocalPart();
	}

	@Override
	public Set<String> getMimeTypes() {
		return Sets.newHashSet("application/xml");
	}

	@Override
	public Set<String> getExtensions() {
		return Sets.newHashSet("xml");
	}


	@Override
	public void triplify(Properties properties, FacadeXGraphBuilder builder) throws IOException, TriplifierHTTPException {
		List<String> xpaths = Triplifier.getPropertyValues(properties, PROPERTY_XPATH);
		if(!xpaths.isEmpty()){
			transformWithXPath(xpaths, properties, builder);
		}else{
			transformSAX(properties, builder);
		}
	}

	@Override
	public Iterable<Slice> slice(Properties properties) throws IOException, TriplifierHTTPException {
		final String dataSourceId = Triplifier.getRootArgument(properties);
		final String root = Triplifier.getRootArgument(properties);
		List<String> xpaths = Triplifier.getPropertyValues(properties, PROPERTY_XPATH);

		try {
			VTDNav vn = buildVTDNav(properties);
			Iterator<String> xit = xpaths.iterator();
			final Iterator<Pair<VTDNav, Integer>> it = evaluateXPaths(vn, xpaths);
			return new Iterable<Slice>() {
				@Override
				public Iterator<Slice> iterator() {

					return new Iterator<Slice>() {
						int theCount = 1;

						@Override
						public boolean hasNext() {
							return it.hasNext();
						}

						@Override
						public Slice next() {
							Pair<VTDNav, Integer> pair = it.next();
							int c = theCount;
							theCount++;
							return XPathSlice.make(pair.getKey(), pair.getValue(), c, root, dataSourceId);
						}
					};
				}
			};
		}catch (Exception e){
			throw new RuntimeException((e));
		}
	}

	@Override
	public void triplify(Slice slice, Properties properties, FacadeXGraphBuilder builder) {
		builder.addRoot(slice.getDatasourceId(), slice.getRootId());
		if(slice instanceof XPathSlice) {
			XPathSlice xs = (XPathSlice) slice;
			try {
				transformFromXPath(xs.get().getKey(), xs.get().getValue(), xs.iteration(), slice.getRootId(), slice.getDatasourceId(),
				properties,  builder);
			} catch (NavException e) {
				throw new RuntimeException(e);
			}
		}else {
			throw new RuntimeException("Not the expected slice (" + XPathSlice.class.toString() + ")");
		}
	}
}
